@@ -54,7 +54,7 @@ struct n800_s {
     DeviceState *tsc;
 
     int keymap[0x80];
-    i2c_slave *kbd;
+    DeviceState *kbd;
 
     DeviceState *usb;
     void *retu;
@@ -172,13 +172,14 @@ static const uint8_t n8x0_cal_bt_id[] = {
 static void n8x0_nand_setup(struct n800_s *s)
 {
     char *otp_region;
+    DriveInfo *dinfo;
 
+    dinfo = drive_get(IF_MTD, 0, 0);
     /* Either 0x40 or 0x48 are OK for the device ID */
-    DriveInfo *dinfo = drive_get(IF_MTD, 0, 0);
-    s->nand = onenand_create(NAND_MFR_SAMSUNG, 0x48, 0, 1,
-                             qdev_get_gpio_in(s->cpu->gpio, N8X0_ONENAND_GPIO),
-                             dinfo ? dinfo->bdrv : NULL);
-    omap_gpmc_attach(s->cpu->gpmc, N8X0_ONENAND_CS, s->nand, 0, 0);
+    s->nand = onenand_init(dinfo ? dinfo->bdrv : 0,
+                           NAND_MFR_SAMSUNG, 0x48, 0, 1,
+                           qdev_get_gpio_in(s->cpu->gpio, N8X0_ONENAND_GPIO));
+    omap_gpmc_attach(s->cpu->gpmc, N8X0_ONENAND_CS, s->nand, 0, OMAP_GPMC_NOR);
     otp_region = onenand_raw_otp(s->nand);
 
     memcpy(otp_region + 0x000, n8x0_cal_wlan_mac, sizeof(n8x0_cal_wlan_mac));
@@ -366,7 +367,6 @@ static int n810_keys[0x80] = {
 static void n810_kbd_setup(struct n800_s *s)
 {
     qemu_irq kbd_irq = qdev_get_gpio_in(s->cpu->gpio, N810_KEYBOARD_GPIO);
-    DeviceState *dev;
     int i;
 
     for (i = 0; i < 0x80; i ++)
@@ -379,10 +379,9 @@ static void n810_kbd_setup(struct n800_s *s)
 
     /* Attach the LM8322 keyboard to the I2C bus,
      * should happen in n8x0_i2c_setup and s->kbd be initialised here.  */
-    dev = i2c_create_slave(omap_i2c_bus(s->cpu->i2c, 0),
+    s->kbd = i2c_create_slave(omap_i2c_bus(s->cpu->i2c, 0),
                            "lm8323", N810_LM8323_ADDR);
-    qdev_connect_gpio_out(dev, 0, kbd_irq);
-    s->kbd = (void *)dev;
+    qdev_connect_gpio_out(s->kbd, 0, kbd_irq);
 }
 
 /* LCD MIPI DBI-C controller (URAL) */
@@ -885,9 +884,9 @@ static void n8x0_uart_setup(struct n800_s *s)
                     qdev_get_gpio_in(s->cpu->gpio, N8X0_BT_HOST_WKUP_GPIO));
 
     qdev_connect_gpio_out(s->cpu->gpio, N8X0_BT_RESET_GPIO,
-                          csrhci_pins_get(radio)[csrhci_pin_reset]);
+                    csrhci_pins_get(radio)[csrhci_pin_reset]);
     qdev_connect_gpio_out(s->cpu->gpio, N8X0_BT_WKUP_GPIO,
-                          csrhci_pins_get(radio)[csrhci_pin_wakeup]);
+                    csrhci_pins_get(radio)[csrhci_pin_wakeup]);
 
     omap_uart_attach(s->cpu->uart[BT_UART], radio, "bt-uart");
 }
@@ -898,10 +897,10 @@ static void n8x0_usb_setup(struct n800_s *s)
     sysbus_connect_irq(sysbus_from_qdev(s->usb), 0,
                        qdev_get_gpio_in(s->cpu->gpio, N8X0_TUSB_INT_GPIO));
     /* Using the NOR interface */
-    omap_gpmc_attach(s->cpu->gpmc, N8X0_USB_SYNC_CS, s->usb, 0, 0);
-    omap_gpmc_attach(s->cpu->gpmc, N8X0_USB_ASYNC_CS, s->usb, 1, 0);
+    omap_gpmc_attach(s->cpu->gpmc, N8X0_USB_SYNC_CS, s->usb, 0, OMAP_GPMC_NOR);
+    omap_gpmc_attach(s->cpu->gpmc, N8X0_USB_ASYNC_CS, s->usb, 1, OMAP_GPMC_NOR);
     qdev_connect_gpio_out(s->cpu->gpio, N8X0_TUSB_ENABLE_GPIO,
-                          qdev_get_gpio_in(s->usb, 0)); // tusb_power
+                          qdev_get_gpio_in(s->usb, 0)); /* tusb_power */
 }
 
 /* Setup done before the main bootloader starts by some early setup code
@@ -1381,12 +1380,12 @@ static int n8x0_atag_setup(void *p, int model)
     return (void *) w - p;
 }
 
-static int n800_atag_setup(struct arm_boot_info *info, void *p)
+static int n800_atag_setup(const struct arm_boot_info *info, void *p)
 {
     return n8x0_atag_setup(p, 800);
 }
 
-static int n810_atag_setup(struct arm_boot_info *info, void *p)
+static int n810_atag_setup(const struct arm_boot_info *info, void *p)
 {
     return n8x0_atag_setup(p, 810);
 }
@@ -1732,7 +1731,7 @@ static void lis302dl_trigger(LIS302DLState *s, int axis, int value)
     lis302dl_interrupt_update(s);
 }
 
-void lis302dl_step(void *opaque, int axis, int high, int activate)
+static void lis302dl_step(void *opaque, int axis, int high, int activate)
 {
     TRACE_LIS302DL("axis=%d, high=%d, activate=%d", axis, high, activate);
     LIS302DLState *s = opaque;
@@ -2503,9 +2502,9 @@ static void n900_init(ram_addr_t ram_size,
     qdev_prop_set_uint8(s->mipid, "n900", 1);
     qdev_init_nofail(s->mipid);
 
-    s->nand = onenand_create(NAND_MFR_SAMSUNG, 0x40, 0x121, 1, 
-                             qdev_get_gpio_in(s->cpu->gpio, N900_ONENAND_GPIO),
-                             dmtd ? dmtd->bdrv : NULL);
+    s->nand = onenand_init(dmtd ? dmtd->bdrv : NULL,
+                           NAND_MFR_SAMSUNG, 0x40, 0x121, 1,
+                           qdev_get_gpio_in(s->cpu->gpio, N900_ONENAND_GPIO));
     omap_gpmc_attach(s->cpu->gpmc, 0, s->nand, 0, 0);
 
     if (dsd) {
