@@ -107,6 +107,9 @@ struct omap_gpmc_s *omap_gpmc_init(struct omap_mpu_state_s *mpu,
 void omap_gpmc_reset(struct omap_gpmc_s *s);
 void omap_gpmc_attach(struct omap_gpmc_s *s, int cs, DeviceState *dev,
                       int mmio_index, int devicetype);
+/* devicetype values for omap_gpmc_attach */
+#define OMAP_GPMC_NOR 0
+#define OMAP_GPMC_NAND 2
 
 /* omap_l4.c */
 struct omap_l4_s;
@@ -158,7 +161,8 @@ struct omap_target_agent_s *omap3_l4ta_init(
     int cs);
 target_phys_addr_t omap_l4_attach(struct omap_target_agent_s *ta, int region,
                                   int iotype);
-target_phys_addr_t omap_l4_base(struct omap_target_agent_s *ta, int region);
+target_phys_addr_t omap_l4_region_base(struct omap_target_agent_s *ta,
+                                       int region);
 uint32_t omap_l4_size(struct omap_target_agent_s *ta, int region);
 int l4_register_io_memory(CPUReadMemoryFunc * const *mem_read,
                           CPUWriteMemoryFunc * const *mem_write,
@@ -992,7 +996,7 @@ SPIBus *omap_mcspi_bus(DeviceState *omap_mcspi, int bus_number);
 # define cpu_is_omap2420(cpu)		(cpu->mpu_model == omap2420)
 # define cpu_is_omap2430(cpu)		(cpu->mpu_model == omap2430)
 # define cpu_is_omap3430(cpu)		(cpu->mpu_model == omap3430)
-# define cpu_is_omap3630(cpu)       (cpu->mpu_model == omap3630)
+# define cpu_is_omap3630(cpu)           (cpu->mpu_model == omap3630)
 
 # define cpu_is_omap15xx(cpu)		\
         (cpu_is_omap310(cpu) || cpu_is_omap1510(cpu))
@@ -1004,7 +1008,7 @@ SPIBus *omap_mcspi_bus(DeviceState *omap_mcspi, int bus_number);
 # define cpu_class_omap1(cpu)		\
         (cpu_is_omap15xx(cpu) || cpu_is_omap16xx(cpu))
 # define cpu_class_omap2(cpu)		cpu_is_omap24xx(cpu)
-# define cpu_class_omap3(cpu)		\
+# define cpu_class_omap3(cpu) \
         (cpu_is_omap3430(cpu) || cpu_is_omap3630(cpu))
 
 struct omap_mpu_state_s {
@@ -1043,7 +1047,9 @@ struct omap_mpu_state_s {
 
     /* MPUI-TIPB peripherals */
     DeviceState *uart[4];
+
     DeviceState *gpio;
+
     struct omap_mcbsp_s *mcbsp1;
     struct omap_mcbsp_s *mcbsp3;
 
@@ -1120,6 +1126,7 @@ struct omap_mpu_state_s {
     DeviceState *omap3_mmc[3];
     DeviceState *omap3_usb_otg;
     DeviceState *omap3_usb_host;
+    DeviceState *omap3_usb_ohci;
     ram_addr_t bootrom_base;
 };
 
