@@ -837,7 +837,7 @@ static struct clk i2c2_iclk = {
     .parent	= &core_l4_iclk,
 };
 
-static struct clk gpio_dbclk[4] = {
+static struct clk gpio_dbclk[5] = {
     {
         .name	= "gpio1_dbclk",
         .flags	= CLOCK_IN_OMAP242X | CLOCK_IN_OMAP243X,
@@ -854,6 +854,10 @@ static struct clk gpio_dbclk[4] = {
         .name	= "gpio4_dbclk",
         .flags	= CLOCK_IN_OMAP242X | CLOCK_IN_OMAP243X,
         .parent	= &wu_32k_clk,
+    }, {
+        .name   = "gpio5_dbclk",
+        .flags  = CLOCK_IN_OMAP243X,
+        .parent = &wu_32k_clk,
     },
 };
 
@@ -1340,6 +1344,12 @@ static struct clk omap3_per_48m_fclk = {
     .parent	= &omap3_48m_fclk,        
 };
 
+static struct clk omap3_per_32k_fclk = {
+    .name	= "omap3_per_32k_fclk",
+    .flags	= CLOCK_IN_OMAP34XX | CLOCK_IN_OMAP36XX,
+    .parent	= &omap3_32k_fclk,
+};
+
 static struct clk omap3_per_l4_iclk = {
     .name	= "omap3_per_l4_iclk",
     .flags	= CLOCK_IN_OMAP34XX | CLOCK_IN_OMAP36XX,
@@ -1759,6 +1769,7 @@ static struct clk *onchip_clks[] = {
     &omap3_gp9_fclk,
     &omap3_per_96m_fclk,
     &omap3_per_48m_fclk,
+    &omap3_per_32k_fclk,
     &omap3_per_l4_iclk,
     &omap3_uart1_fclk,
     &omap3_uart1_iclk,
