@@ -57,6 +57,19 @@ struct omap_gpmc_s {
     } prefetch;
 };
 
+#define OMAP_GPMC_8BIT 0
+#define OMAP_GPMC_16BIT 1
+
+static int omap_gpmc_devtype(struct omap_gpmc_cs_file_s *f)
+{
+    return (f->config[0] >> 10) & 3;
+}
+
+static int omap_gpmc_devsize(struct omap_gpmc_cs_file_s *f)
+{
+    return (f->config[0] >> 12) & 3;
+}
+
 /* Extract the chip-select value from the prefetch config1 register */
 static int prefetch_cs(uint32_t config1)
 {
@@ -103,10 +116,10 @@ static uint32_t omap_nand_read8(void *opaque, target_phys_addr_t addr)
 {
     struct omap_gpmc_cs_file_s *f = (struct omap_gpmc_cs_file_s *)opaque;
     nand_setpins(f->dev, 0, 0, 0, 1, 0);
-    switch (((f->config[0] >> 12) & 3)) {
-    case 0: /* 8bit */
+    switch (omap_gpmc_devsize(f)) {
+    case OMAP_GPMC_8BIT:
         return nand_getio(f->dev);
-    case 1: /* 16bit */
+    case OMAP_GPMC_16BIT:
         /* reading 8bits from a 16bit device?! */
         return nand_getio(f->dev);
     default:
@@ -119,12 +132,12 @@ static uint32_t omap_nand_read16(void *opaque, target_phys_addr_t addr)
     struct omap_gpmc_cs_file_s *f = (struct omap_gpmc_cs_file_s *)opaque;
     uint32_t x1, x2;
     nand_setpins(f->dev, 0, 0, 0, 1, 0);
-    switch (((f->config[0] >> 12) & 3)) {
-    case 0: /* 8bit */
+    switch (omap_gpmc_devsize(f)) {
+    case OMAP_GPMC_8BIT:
         x1 = nand_getio(f->dev);
         x2 = nand_getio(f->dev);
         return (x2 << 8) | x1;
-    case 1: /* 16bit */
+    case OMAP_GPMC_16BIT:
         return nand_getio(f->dev);
     default:
         return 0;
@@ -136,14 +149,14 @@ static uint32_t omap_nand_read32(void *opaque, target_phys_addr_t addr)
     struct omap_gpmc_cs_file_s *f = (struct omap_gpmc_cs_file_s *)opaque;
     uint32_t x1, x2, x3, x4;
     nand_setpins(f->dev, 0, 0, 0, 1, 0);
-    switch (((f->config[0] >> 12) & 3)) {
-    case 0: /* 8bit */
+    switch (omap_gpmc_devsize(f)) {
+    case OMAP_GPMC_8BIT:
         x1 = nand_getio(f->dev);
         x2 = nand_getio(f->dev);
         x3 = nand_getio(f->dev);
         x4 = nand_getio(f->dev);
         return (x4 << 24) | (x3 << 16) | (x2 << 8) | x1;
-    case 1: /* 16bit */
+    case OMAP_GPMC_16BIT:
         x1 = nand_getio(f->dev);
         x2 = nand_getio(f->dev);
         return (x2 << 16) | x1;
@@ -152,15 +165,16 @@ static uint32_t omap_nand_read32(void *opaque, target_phys_addr_t addr)
     }
 }
 
-static void omap_nand_write8(void *opaque, target_phys_addr_t addr, uint32_t value)
+static void omap_nand_write8(void *opaque, target_phys_addr_t addr,
+                             uint32_t value)
 {
     struct omap_gpmc_cs_file_s *f = (struct omap_gpmc_cs_file_s *)opaque;
     nand_setpins(f->dev, 0, 0, 0, 1, 0);
-    switch (((f->config[0] >> 12) & 3)) {
-    case 0: /* 8bit */
+    switch (omap_gpmc_devsize(f)) {
+    case OMAP_GPMC_8BIT:
         nand_setio(f->dev, value & 0xff);
         break;
-    case 1: /* 16bit */
+    case OMAP_GPMC_16BIT:
         /* writing to a 16bit device with 8bit access!? */
         nand_setio(f->dev, value & 0xffff);
         break;
@@ -169,16 +183,17 @@ static void omap_nand_write8(void *opaque, target_phys_addr_t addr, uint32_t val
     }
 }
 
-static void omap_nand_write16(void *opaque, target_phys_addr_t addr, uint32_t value)
+static void omap_nand_write16(void *opaque, target_phys_addr_t addr,
+                              uint32_t value)
 {
     struct omap_gpmc_cs_file_s *f = (struct omap_gpmc_cs_file_s *)opaque;
     nand_setpins(f->dev, 0, 0, 0, 1, 0);
-    switch (((f->config[0] >> 12) & 3)) {
-    case 0: /* 8bit */
+    switch (omap_gpmc_devsize(f)) {
+    case OMAP_GPMC_8BIT:
         nand_setio(f->dev, value & 0xff);
         nand_setio(f->dev, (value >> 8) & 0xff);
         break;
-    case 1: /* 16bit */
+    case OMAP_GPMC_16BIT:
         nand_setio(f->dev, value & 0xffff);
         break;
     default:
@@ -186,18 +201,19 @@ static void omap_nand_write16(void *opaque, target_phys_addr_t addr, uint32_t va
     }
 }
 
-static void omap_nand_write32(void *opaque, target_phys_addr_t addr, uint32_t value)
+static void omap_nand_write32(void *opaque, target_phys_addr_t addr,
+                              uint32_t value)
 {
     struct omap_gpmc_cs_file_s *f = (struct omap_gpmc_cs_file_s *)opaque;
     nand_setpins(f->dev, 0, 0, 0, 1, 0);
-    switch (((f->config[0] >> 12) & 3)) {
-    case 0: /* 8bit */
+    switch (omap_gpmc_devsize(f)) {
+    case OMAP_GPMC_8BIT:
         nand_setio(f->dev, value & 0xff);
         nand_setio(f->dev, (value >> 8) & 0xff);
         nand_setio(f->dev, (value >> 16) & 0xff);
         nand_setio(f->dev, (value >> 24) & 0xff);
         break;
-    case 1: /* 16bit */
+    case OMAP_GPMC_16BIT:
         nand_setio(f->dev, value & 0xffff);
         nand_setio(f->dev, (value >> 16) & 0xffff);
         break;
@@ -206,15 +222,13 @@ static void omap_nand_write32(void *opaque, target_phys_addr_t addr, uint32_t va
     }
 }
 
-static CPUReadMemoryFunc *omap_nand_readfn[] =
-{
+static CPUReadMemoryFunc *omap_nand_readfn[] = {
     omap_nand_read8,
     omap_nand_read16,
     omap_nand_read32,
 };
 
-static CPUWriteMemoryFunc *omap_nand_writefn[] =
-{
+static CPUWriteMemoryFunc *omap_nand_writefn[] = {
     omap_nand_write8,
     omap_nand_write16,
     omap_nand_write32,
@@ -304,7 +318,7 @@ static uint32_t omap_gpmc_prefetch_read8(void *opaque, target_phys_addr_t addr)
     }
     data = s->prefetch.fifo[63 - s->prefetch.fifopointer];
     if (s->prefetch.fifopointer ==
-        (64 - prefetch_threshold (s->prefetch.config1))) {
+        (64 - prefetch_threshold(s->prefetch.config1))) {
         /* We've drained THRESHOLD bytes now. So deassert the
          * DMA request, then refill the FIFO (which will probably
          * assert it again.)
@@ -394,15 +408,13 @@ static void omap_gpmc_prefetch_write32(void *opaque, target_phys_addr_t addr,
     omap_gpmc_prefetch_write8(opaque, addr, (value >> 24) & 0xff);
 }
 
-static CPUReadMemoryFunc *omap_gpmc_prefetch_readfn[] =
-{
+static CPUReadMemoryFunc *omap_gpmc_prefetch_readfn[] = {
     omap_gpmc_prefetch_read8,
     omap_gpmc_prefetch_read16,
     omap_gpmc_prefetch_read32,
 };
 
-static CPUWriteMemoryFunc *omap_gpmc_prefetch_writefn[] =
-{
+static CPUWriteMemoryFunc *omap_gpmc_prefetch_writefn[] = {
     omap_gpmc_prefetch_write8,
     omap_gpmc_prefetch_write16,
     omap_gpmc_prefetch_write32,
@@ -411,24 +423,23 @@ static CPUWriteMemoryFunc *omap_gpmc_prefetch_writefn[] =
 static void omap_gpmc_cs_map(struct omap_gpmc_s *s, int cs)
 {
     struct omap_gpmc_cs_file_s *f = &s->cs_file[cs];
+    uint32_t mask = (f->config[6] >> 8) & 0xf;
+    uint32_t base = f->config[6] & 0x3f;
+    uint32_t size;
     if (!(f->config[6] & (1 << 6))) {
         /* Do nothing unless CSVALID */
         return;
     }
-
-    uint32_t mask = (f->config[6] >> 8) & 0xf;
-    uint32_t base = f->config[6] & 0x3f;
-    uint32_t size;
     /* TODO: check for overlapping regions and report access errors */
     if (mask != 0x8 && mask != 0xc && mask != 0xe && mask != 0xf
          && !(s->accept_256 && !mask)) {
         fprintf(stderr, "%s: invalid chip-select mask address (0x%x)\n",
-                 __FUNCTION__, mask);
+                 __func__, mask);
     }
     base <<= 24;
     size = (0x0fffffff & ~(mask << 24)) + 1;
-    switch ((f->config[0] >> 10) & 3) {
-    case 0: /* DEVICETYPE == NOR */
+    switch (omap_gpmc_devtype(f)) {
+    case OMAP_GPMC_NOR:
         /* TODO: rather than setting the size of the mapping (which should be
          * constant), the mask should cause wrapping of the address space, so
          * that the same memory becomes accessible at every <i>size</i> bytes
@@ -438,7 +449,7 @@ static void omap_gpmc_cs_map(struct omap_gpmc_s *s, int cs)
             sysbus_mmio_map(sysbus_from_qdev(f->dev), f->mmio_index, base);
         }
         break;
-    case 2: /* DEVICETYPE == NAND */
+    case OMAP_GPMC_NAND:
         if ((s->prefetch.config1 & 0x80) &&
             (prefetch_cs(s->prefetch.config1) == cs)) {
             /* The prefetch engine is enabled for this CS: map the FIFO */
@@ -460,13 +471,13 @@ static void omap_gpmc_cs_unmap(struct omap_gpmc_s *s, int cs)
         return;
     }
 
-    switch ((f->config[0] >> 10) & 3) {
-    case 0: /* DEVICETYPE == NOR */
+    switch (omap_gpmc_devtype(f)) {
+    case OMAP_GPMC_NOR:
         if (f->dev && f->mmio_index >= 0) {
             sysbus_mmio_unmap(sysbus_from_qdev(f->dev), f->mmio_index);
         }
         break;
-    case 2: /* DEVICETYPE == NAND */
+    case OMAP_GPMC_NAND:
         {
             uint32_t mask = (f->config[6] >> 8) & 0xf;
             uint32_t base = (f->config[6] & 0x3f) << 24;
@@ -496,10 +507,10 @@ void omap_gpmc_reset(struct omap_gpmc_s *s)
     s->prefetch.count = 0;
     for (i = 0; i < 8; i ++) {
         omap_gpmc_cs_unmap(s, i);
-        s->cs_file[i].config[1] = 0x00101001;
-        s->cs_file[i].config[2] = 0x00020201;
+        s->cs_file[i].config[1] = 0x101001;
+        s->cs_file[i].config[2] = 0x020201;
         s->cs_file[i].config[3] = 0x10031003;
-        s->cs_file[i].config[4] = 0x010f1111;
+        s->cs_file[i].config[4] = 0x10f1111;
         s->cs_file[i].config[5] = 0;
         s->cs_file[i].config[6] = 0xf00;
         /* FIXME: attached devices should be probed for some of the CFG1 bits
@@ -518,9 +529,8 @@ void omap_gpmc_reset(struct omap_gpmc_s *s)
     s->ecc_cs = 0;
     s->ecc_ptr = 0;
     s->ecc_cfg = 0x3fcff000;
-    for (i = 0; i < 9; i ++) {
+    for (i = 0; i < 9; i ++)
         ecc_reset(&s->ecc[i]);
-    }
 }
 
 static uint32_t omap_gpmc_read32(void *opaque, target_phys_addr_t addr)
@@ -563,27 +573,25 @@ static uint32_t omap_gpmc_read32(void *opaque, target_phys_addr_t addr)
         addr -= cs * 0x30;
         f = s->cs_file + cs;
         switch (addr) {
-        case 0x60:	/* GPMC_CONFIG1 */
+        case 0x60:      /* GPMC_CONFIG1 */
             return f->config[0];
-        case 0x64:	/* GPMC_CONFIG2 */
+        case 0x64:      /* GPMC_CONFIG2 */
             return f->config[1];
-        case 0x68:	/* GPMC_CONFIG3 */
+        case 0x68:      /* GPMC_CONFIG3 */
             return f->config[2];
-        case 0x6c:	/* GPMC_CONFIG4 */
+        case 0x6c:      /* GPMC_CONFIG4 */
             return f->config[3];
-        case 0x70:	/* GPMC_CONFIG5 */
+        case 0x70:      /* GPMC_CONFIG5 */
             return f->config[4];
-        case 0x74:	/* GPMC_CONFIG6 */
+        case 0x74:      /* GPMC_CONFIG6 */
             return f->config[5];
-        case 0x78:	/* GPMC_CONFIG7 */
+        case 0x78:      /* GPMC_CONFIG7 */
             return f->config[6];
-        case 0x84:	/* GPMC_NAND_DATA */
-            if (((f->config[0] >> 10) & 3) == 2) { /* NAND like? */
+        case 0x84:      /* GPMC_NAND_DATA */
+            if (omap_gpmc_devtype(f) == OMAP_GPMC_NAND) {
                 return omap_nand_read32(f, 0);
             }
             return 0;
-        default:
-            break;
         }
         break;
 
@@ -608,7 +616,8 @@ static uint32_t omap_gpmc_read32(void *opaque, target_phys_addr_t addr)
     case 0x200 ... 0x220:	/* GPMC_ECC_RESULT */
         cs = (addr & 0x1f) >> 2;
         /* TODO: check correctness */
-        return ((s->ecc[cs].cp    &  0x07) <<  0) |
+        return
+                ((s->ecc[cs].cp    &  0x07) <<  0) |
                 ((s->ecc[cs].cp    &  0x38) << 13) |
                 ((s->ecc[cs].lp[0] & 0x1ff) <<  3) |
                 ((s->ecc[cs].lp[1] & 0x1ff) << 19);
@@ -636,13 +645,13 @@ static uint32_t omap_gpmc_read8(void *opaque, target_phys_addr_t addr)
         addr -= cs * 0x30;
         f = s->cs_file + cs;
         switch (addr) {
-        case 0x84 ... 0x87:	/* GPMC_NAND_DATA */
-            if (((f->config[0] >> 10) & 3) == 2) { /* NAND like? */
+        case 0x84 ... 0x87: /* GPMC_NAND_DATA */
+            if (omap_gpmc_devtype(f) == OMAP_GPMC_NAND) {
                 return omap_nand_read8(f, 0);
             }
             return 0;
         default:
-            return 0;
+            break;
         }
         break;
     default:
@@ -664,9 +673,9 @@ static uint32_t omap_gpmc_read16(void *opaque, target_phys_addr_t addr)
         addr -= cs * 0x30;
         f = s->cs_file + cs;
         switch (addr) {
-        case 0x84:	/* GPMC_NAND_DATA */
+        case 0x84: /* GPMC_NAND_DATA */
         case 0x86:
-            if (((f->config[0] >> 10) & 3) == 2) { /* NAND like? */
+            if (omap_gpmc_devtype(f) == OMAP_GPMC_NAND) {
                 return omap_nand_read16(f, 0);
             }
             return 0;
@@ -702,7 +711,7 @@ static void omap_gpmc_write32(void *opaque, target_phys_addr_t addr,
     case 0x010:	/* GPMC_SYSCONFIG */
         if ((value >> 3) == 0x3)
             fprintf(stderr, "%s: bad SDRAM idle mode %i\n",
-                    __FUNCTION__, value >> 3);
+                            __FUNCTION__, value >> 3);
         if (value & 2)
             omap_gpmc_reset(s);
         s->sysconfig = value & 0x19;
@@ -735,47 +744,52 @@ static void omap_gpmc_write32(void *opaque, target_phys_addr_t addr,
         addr -= cs * 0x30;
         f = s->cs_file + cs;
         switch (addr) {
-        case 0x60:	/* GPMC_CONFIG1 */
+        case 0x60:      /* GPMC_CONFIG1 */
             f->config[0] = value & 0xffef3e13;
             break;
-        case 0x64:	/* GPMC_CONFIG2 */
+        case 0x64:      /* GPMC_CONFIG2 */
             f->config[1] = value & 0x001f1f8f;
             break;
-        case 0x68:	/* GPMC_CONFIG3 */
+        case 0x68:      /* GPMC_CONFIG3 */
             f->config[2] = value & 0x001f1f8f;
             break;
-        case 0x6c:	/* GPMC_CONFIG4 */
+        case 0x6c:      /* GPMC_CONFIG4 */
             f->config[3] = value & 0x1f8f1f8f;
             break;
-        case 0x70:	/* GPMC_CONFIG5 */
+        case 0x70:      /* GPMC_CONFIG5 */
             f->config[4] = value & 0x0f1f1f1f;
             break;
-        case 0x74:	/* GPMC_CONFIG6 */
+        case 0x74:      /* GPMC_CONFIG6 */
             f->config[5] = value & 0x00000fcf;
             break;
-        case 0x78:	/* GPMC_CONFIG7 */
+        case 0x78:      /* GPMC_CONFIG7 */
             if ((f->config[6] ^ value) & 0xf7f) {
                 omap_gpmc_cs_unmap(s, cs);
                 f->config[6] = value & 0x00000f7f;
                 omap_gpmc_cs_map(s, cs);
             }
             break;
-        case 0x7c:	/* GPMC_NAND_COMMAND */
-        case 0x80:	/* GPMC_NAND_ADDRESS */
-            if (((f->config[0] >> 10) & 3) == 2) { /* NAND like? */
+        case 0x7c:      /* GPMC_NAND_COMMAND */
+        case 0x80:      /* GPMC_NAND_ADDRESS */
+            if (omap_gpmc_devtype(f) == OMAP_GPMC_NAND) {
                 switch (addr) {
-                case 0x7c: nand_setpins(f->dev, 1, 0, 0, 1, 0); break; /* CLE */
-                case 0x80: nand_setpins(f->dev, 0, 1, 0, 1, 0); break; /* ALE */
-                default: break;
+                case 0x7c: /* CLE */
+                    nand_setpins(f->dev, 1, 0, 0, 1, 0);
+                    break;
+                case 0x80: /* ALE */
+                    nand_setpins(f->dev, 0, 1, 0, 1, 0);
+                    break;
+                default:
+                    break;
                 }
-                switch (((f->config[0] >> 12) & 3)) {
-                case 0: /* 8bit */
+                switch (omap_gpmc_devsize(f)) {
+                case OMAP_GPMC_8BIT:
                     nand_setio(f->dev, value & 0xff);
                     nand_setio(f->dev, (value >> 8) & 0xff);
                     nand_setio(f->dev, (value >> 16) & 0xff);
                     nand_setio(f->dev, (value >> 24) & 0xff);
                     break;
-                case 1: /* 16bit */
+                case OMAP_GPMC_16BIT:
                     nand_setio(f->dev, value & 0xffff);
                     nand_setio(f->dev, (value >> 16) & 0xffff);
                     break;
@@ -785,7 +799,7 @@ static void omap_gpmc_write32(void *opaque, target_phys_addr_t addr,
             }
             break;
         case 0x84:  /* GPMC_NAND_DATA */
-            if (((f->config[0] >> 10) & 3) == 2) { /* NAND like? */
+            if (omap_gpmc_devtype(f) == OMAP_GPMC_NAND) {
                 omap_nand_write32(f, 0, value);
             }
             break;
@@ -893,9 +907,9 @@ static void omap_gpmc_write8(void *opaque, target_phys_addr_t addr,
         addr -= cs * 0x30;
         f = s->cs_file + cs;
         switch (addr) {
-        case 0x7c ... 0x7f:	/* GPMC_NAND_COMMAND */
-        case 0x80 ... 0x83:	/* GPMC_NAND_ADDRESS */
-            if (((f->config[0] >> 10) & 3) == 2) { /* NAND like? */
+        case 0x7c ... 0x7f:     /* GPMC_NAND_COMMAND */
+        case 0x80 ... 0x83:     /* GPMC_NAND_ADDRESS */
+            if (omap_gpmc_devtype(f) == OMAP_GPMC_NAND) {
                 switch (addr) {
                 case 0x7c ... 0x7f:
                     nand_setpins(f->dev, 1, 0, 0, 1, 0); /* CLE */
@@ -906,11 +920,11 @@ static void omap_gpmc_write8(void *opaque, target_phys_addr_t addr,
                 default:
                     break;
                 }
-                switch (((f->config[0] >> 12) & 3)) {
-                case 0: /* 8bit */
+                switch (omap_gpmc_devsize(f)) {
+                case OMAP_GPMC_8BIT:
                     nand_setio(f->dev, value & 0xff);
                     break;
-                case 1: /* 16bit */
+                case OMAP_GPMC_16BIT:
                     /* writing to a 16bit device with 8bit access!? */
                     nand_setio(f->dev, value & 0xffff);
                     break;
@@ -920,7 +934,7 @@ static void omap_gpmc_write8(void *opaque, target_phys_addr_t addr,
             }
             break;
         case 0x84 ... 0x87: /* GPMC_NAND_DATA */
-            if (((f->config[0] >> 10) & 3) == 2) { /* NAND like? */
+            if (omap_gpmc_devtype(f) == OMAP_GPMC_NAND) {
                 omap_nand_write8(f, 0, value);
             }
             break;
@@ -948,14 +962,14 @@ static void omap_gpmc_write16(void *opaque, target_phys_addr_t addr,
         addr -= cs * 0x30;
         f = s->cs_file + cs;
         switch (addr) {
-        case 0x7c:	/* GPMC_NAND_COMMAND */
+        case 0x7c:      /* GPMC_NAND_COMMAND */
         case 0x7e:
-        case 0x80:	/* GPMC_NAND_ADDRESS */
+        case 0x80:      /* GPMC_NAND_ADDRESS */
         case 0x82:
-            if (((f->config[0] >> 10) & 3) == 2) { /* NAND like? */
+            if (omap_gpmc_devtype(f) == OMAP_GPMC_NAND) {
                 switch (addr) {
                 case 0x7c:
-                case 0x7e: 
+                case 0x7e:
                     nand_setpins(f->dev, 1, 0, 0, 1, 0); /* CLE */
                     break;
                 case 0x80:
@@ -969,12 +983,12 @@ static void omap_gpmc_write16(void *opaque, target_phys_addr_t addr,
                 default:
                     break;
                 }
-                switch (((f->config[0] >> 12) & 3)) {
-                case 0: /* 8bit */
+                switch (omap_gpmc_devsize(f)) {
+                case OMAP_GPMC_8BIT:
                     nand_setio(f->dev, value & 0xff);
                     nand_setio(f->dev, (value >> 8) & 0xff);
                     break;
-                case 1: /* 16bit */
+                case OMAP_GPMC_16BIT:
                     nand_setio(f->dev, value & 0xffff);
                     break;
                 default:
@@ -982,9 +996,9 @@ static void omap_gpmc_write16(void *opaque, target_phys_addr_t addr,
                 }
             }
             break;
-        case 0x84:  /* GPMC_NAND_DATA */
+        case 0x84:      /* GPMC_NAND_DATA */
         case 0x86:
-            if (((f->config[0] >> 10) & 3) == 2) { /* NAND like? */
+            if (omap_gpmc_devtype(f) == OMAP_GPMC_NAND) {
                 omap_nand_write16(f, 0, value);
             }
             break;
@@ -1016,7 +1030,8 @@ struct omap_gpmc_s *omap_gpmc_init(struct omap_mpu_state_s *mpu,
                                    qemu_irq irq, qemu_irq drq)
 {
     int iomemtype, cs;
-    struct omap_gpmc_s *s = qemu_mallocz(sizeof(*s));
+    struct omap_gpmc_s *s = (struct omap_gpmc_s *)
+            qemu_mallocz(sizeof(struct omap_gpmc_s));
 
     s->irq = irq;
     s->drq = drq;
@@ -1050,20 +1065,22 @@ void omap_gpmc_attach(struct omap_gpmc_s *s, int cs, DeviceState *dev,
     struct omap_gpmc_cs_file_s *f;
 
     if (cs < 0 || cs >= 8) {
-        hw_error("%s: bad chip-select %i\n", __FUNCTION__, cs);
+        fprintf(stderr, "%s: bad chip-select %i\n", __FUNCTION__, cs);
+        exit(-1);
     }
     f = &s->cs_file[cs];
     if (dev != f->dev || mmio_index != f->mmio_index ||
-        devicetype != ((f->config[0] >> 10) & 3)) {
+        devicetype != omap_gpmc_devtype(f)) {
         omap_gpmc_cs_unmap(s, cs);
         f->dev = dev;
         f->mmio_index = mmio_index;
         f->config[0] &= ~(0x3 << 10);
         f->config[0] |= (devicetype & 3) << 10;
-        if (devicetype == 2) { /* NAND */
-            f->config[0] &= ~(0x3 << 12);
-            if (nand_getbuswidth(f->dev) == 16)
-                f->config[0] |= 1 << 12;
+        f->config[0] &= ~(0x3 << 12);
+        if (omap_gpmc_devtype(f) == OMAP_GPMC_NAND) {
+            if (nand_getbuswidth(f->dev) == 16) {
+                f->config[0] |= OMAP_GPMC_16BIT << 12;
+            }
         }
         omap_gpmc_cs_map(s, cs);
     }
