@@ -445,6 +445,12 @@ static uint32_t omap2_inth_read(void *opaque, target_phys_addr_t addr)
     case 0x9c:	/* INTC_PENDING_FIQ */
         return bank->irqs & ~bank->mask & bank->fiq;
 
+    case 0xf8:
+        /* Linux reads this nonexistent interrupt status register,
+         * so treat it as RAZ, the way the hardware does.
+         */
+        return 0;
+
     /* Per-line registers */
     case 0x100 ... 0x300:	/* INTC_ILR */
         bank_no = (offset - 0x100) >> 7;
