@@ -23,8 +23,6 @@
 #include "flash.h"
 #include "irq.h"
 #include "blockdev.h"
-#include "sysemu.h"
-#include "hw.h"
 #include "sysbus.h"
 
 /* 11 for 2kB-page OneNAND ("2nd generation") and 10 for 1kB-page chips */
@@ -121,7 +119,7 @@ static void onenand_base_update(SysBusDevice *dev, target_phys_addr_t new)
                                          0x0200 << s->shift, s->iomemtype);
             cpu_register_physical_memory(new + (0x0200 << s->shift),
                                          0xbe00 << s->shift,
-                                         (s->ram +(0x0200 << s->shift))
+                                         (s->ram + (0x0200 << s->shift))
                                          | IO_MEM_RAM);
             if (s->iomemtype) {
                 cpu_register_physical_memory_offset(new + (0xc000 << s->shift),
@@ -131,7 +129,7 @@ static void onenand_base_update(SysBusDevice *dev, target_phys_addr_t new)
             }
         }
         s->base = new;
-    }        
+    }
 }
 
 static void onenand_intr_update(OneNANDState *s)
@@ -142,28 +140,29 @@ static void onenand_intr_update(OneNANDState *s)
 static void onenand_pre_save(void *opaque)
 {
     OneNANDState *s = opaque;
-    if (s->current == s->otp)
+    if (s->current == s->otp) {
         s->current_direction = 1;
-    else if (s->current == s->image)
+    } else if (s->current == s->image) {
         s->current_direction = 2;
-    else
+    } else {
         s->current_direction = 0;
+    }
 }
 
 static int onenand_post_load(void *opaque, int version_id)
 {
     OneNANDState *s = opaque;
     switch (s->current_direction) {
-        case 0:
-            break;
-        case 1:
-            s->current = s->otp;
-            break;
-        case 2:
-            s->current = s->image;
-            break;
-        default:
-            return -1;
+    case 0:
+        break;
+    case 1:
+        s->current = s->otp;
+        break;
+    case 2:
+        s->current = s->image;
+        break;
+    default:
+        return -1;
     }
     onenand_intr_update(s);
     return 0;
@@ -217,16 +216,17 @@ static void onenand_reset(OneNANDState *s, int cold)
     s->wpstatus = 0x0002;
     s->cycle = 0;
     s->otpmode = 0;
+    s->bdrv_cur = s->bdrv;
     s->current = s->image;
     s->secs_cur = s->secs;
-    s->bdrv_cur = s->bdrv ? s->bdrv : NULL;
 
     if (cold) {
         /* Lock the whole flash */
         memset(s->blockwp, ONEN_LOCK_LOCKED, s->blocks);
 
-        if (s->bdrv_cur && bdrv_read(s->bdrv_cur, 0, s->boot[0], 8) < 0)
-            hw_error("%s: Loading the BootRAM failed.\n", __FUNCTION__);
+        if (s->bdrv_cur && bdrv_read(s->bdrv_cur, 0, s->boot[0], 8) < 0) {
+            hw_error("%s: Loading the BootRAM failed.\n", __func__);
+        }
     }
 }
 
@@ -345,48 +345,48 @@ static inline int onenand_prog_spare(OneNANDState *s, int sec, int secn,
 
 static inline int onenand_erase(OneNANDState *s, int sec, int num)
 {
-    int result = 0;
-
-    uint8_t *buf, *buf2;
-    buf = qemu_malloc(512);
-    if (buf) {
-        buf2 = qemu_malloc(512);
-        if (buf2) {
-            memset(buf, 0xff, 512);
-            for (; !result && num > 0; num--, sec++) {
-                if (s->bdrv_cur) {
-                    result = bdrv_write(s->bdrv_cur, sec, buf, 1);
-                    if (!result) {
-                        result = bdrv_read(s->bdrv_cur,
-                                           s->secs_cur + (sec >> 5),
-                                           buf2, 1) < 0;
-                        if (!result) {
-                            memcpy(buf2 + ((sec & 31) << 4), buf, 1 << 4);
-                            result = bdrv_write(s->bdrv_cur,
-                                                s->secs_cur + (sec >> 5),
-                                                buf2, 1) < 0;
-                        }
-                    }
-                } else {
-                    if (sec + 1 > s->secs_cur) {
-                        result = 1;
-                    } else {
-                        memcpy(s->current + (sec << 9), buf, 512);
-                        memcpy(s->current + (s->secs_cur << 9) + (sec << 4),
-                               buf, 1 << 4);
-                    }
-                }
+    uint8_t *blankbuf, *tmpbuf;
+    blankbuf = qemu_malloc(512);
+    if (!blankbuf) {
+        return 1;
+    }
+    tmpbuf = qemu_malloc(512);
+    if (!tmpbuf) {
+        qemu_free(blankbuf);
+        return 1;
+    }
+    memset(blankbuf, 0xff, 512);
+    for (; num > 0; num--, sec++) {
+        if (s->bdrv_cur) {
+            int erasesec = s->secs_cur + (sec >> 5);
+            if (bdrv_write(s->bdrv_cur, sec, blankbuf, 1)) {
+                goto fail;
+            }
+            if (bdrv_read(s->bdrv_cur, erasesec, tmpbuf, 1) < 0) {
+                goto fail;
+            }
+            memcpy(tmpbuf + ((sec & 31) << 4), blankbuf, 1 << 4);
+            if (bdrv_write(s->bdrv_cur, erasesec, tmpbuf, 1) < 0) {
+                goto fail;
             }
-            qemu_free(buf2);
         } else {
-            result = 1;
+            if (sec + 1 > s->secs_cur) {
+                goto fail;
+            }
+            memcpy(s->current + (sec << 9), blankbuf, 512);
+            memcpy(s->current + (s->secs_cur << 9) + (sec << 4),
+                   blankbuf, 1 << 4);
         }
-        qemu_free(buf);
-    } else {
-        result = 1;
     }
 
-    return result;
+    qemu_free(tmpbuf);
+    qemu_free(blankbuf);
+    return 0;
+
+fail:
+    qemu_free(tmpbuf);
+    qemu_free(blankbuf);
+    return 1;
 }
 
 static void onenand_command(OneNANDState *s)
@@ -446,7 +446,6 @@ static void onenand_command(OneNANDState *s)
         SETADDR(ONEN_BUF_BLOCK, ONEN_BUF_PAGE)
 
         SETBUF_M()
-
         if (onenand_prog_main(s, sec, s->count, buf))
             s->status |= ONEN_ERR_CMD | ONEN_ERR_PROG;
 
@@ -591,7 +590,7 @@ static void onenand_command(OneNANDState *s)
         s->status |= ONEN_ERR_CMD;
         s->intstatus |= ONEN_INT;
         fprintf(stderr, "%s: unknown OneNAND command %x\n",
-                        __FUNCTION__, s->command);
+                        __func__, s->command);
     }
 
     onenand_intr_update(s);
@@ -776,7 +775,7 @@ static CPUWriteMemoryFunc * const onenand_writefn[] = {
     onenand_write,
 };
 
-static int onenand_init(SysBusDevice *dev)
+static int onenand_initfn(SysBusDevice *dev)
 {
     OneNANDState *s = (OneNANDState *)dev;
     uint32_t size = 1 << (24 + ((s->id.dev >> 4) & 7));
@@ -817,7 +816,7 @@ static int onenand_init(SysBusDevice *dev)
 }
 
 static SysBusDeviceInfo onenand_info = {
-    .init = onenand_init,
+    .init = onenand_initfn,
     .qdev.name = "onenand",
     .qdev.size = sizeof(OneNANDState),
     .qdev.reset = onenand_system_reset,
@@ -836,16 +835,17 @@ static void onenand_register_device(void)
     sysbus_register_withprop(&onenand_info);
 }
 
-DeviceState *onenand_create(uint16_t man_id, uint16_t dev_id, uint16_t ver_id,
-                            int regshift, qemu_irq irq, BlockDriverState *bs)
+DeviceState *onenand_init(BlockDriverState *bdrv,
+                          uint16_t man_id, uint16_t dev_id, uint16_t ver_id,
+                          int regshift, qemu_irq irq)
 {
     DeviceState *dev = qdev_create(NULL, "onenand");
     qdev_prop_set_uint16(dev, "manufacturer_id", man_id);
     qdev_prop_set_uint16(dev, "device_id", dev_id);
     qdev_prop_set_uint16(dev, "version_id", ver_id);
     qdev_prop_set_int32(dev, "shift", regshift);
-    if (bs) {
-        qdev_prop_set_drive_nofail(dev, "drive", bs);
+    if (bdrv) {
+        qdev_prop_set_drive_nofail(dev, "drive", bdrv);
     }
     qdev_init_nofail(dev);
     sysbus_connect_irq(sysbus_from_qdev(dev), 0, irq);
